import json
import openai
from datetime import datetime


from promptflow import tool
from promptflow.runtime import PromptFlowRuntime
from promptflow.connections import AzureOpenAIConnection
from promptflow.azure import PFClient

from langchain.agents import Tool, AgentType, initialize_agent
from langchain.chat_models import AzureChatOpenAI
from langchain.chat_models import AzureChatOpenAI


import semantic_kernel as sk
from semantic_kernel.connectors.ai.open_ai import AzureChatCompletion
from semantic_kernel.orchestration.context_variables import ContextVariables

from azure.identity import DefaultAzureCredential
from azure.ai.ml import MLClient

from flow_agent_package.tools.contracts import AgentSkillConfiguration, ArbitrationMethod
from flow_agent_package.tools.constants import format_response, intent_prompt, intent_prompt_with_history, extract_input
from flow_agent_package.tools.orchestrator import OrchestratorPlugin
from flow_agent_package.tools.memory import MemoryHelper
from flow_agent_package.tools.agent_skill import AgentSkill

class ArbitrationAgent:

  def __init__(self, arbitration_method, agent_connection, skills, history):
    self.arbitration_method = arbitration_method
    self.agent_connection = agent_connection
    self.skills = skills
    self.memory_helper = MemoryHelper(history)

    openai.api_base = agent_connection.api_base
    openai.api_type = agent_connection.api_type
    openai.api_version = "2023-07-01-preview"
    openai.api_key = agent_connection.api_key

  def execute(self, query):
    if self.arbitration_method == ArbitrationMethod.LANGCHAIN.value:
      self.use_langchain(query)
    elif self.arbitration_method == ArbitrationMethod.OPENAI_FUNCTIONS.value:
      self.use_openai_functions(query)
    elif self.arbitration_method == ArbitrationMethod.SEMANTIC_KERNEL.value:
      self.use_semantic_kernel(query)
    else:
      raise Exception(f"Invalid arbitration method used: {self.arbitration_method}")
  
  def use_langchain(self, query):
    print("Using Langchain Arbitration")
    tools = []
    for skill in self.skills:
      run_func = skill.to_langchain_function()
      lang_tool = Tool(
        name=skill.name,
        func=run_func,
        description=skill.tool_description,
        return_direct=True
      )
      tools.append(lang_tool)
    
    # todo: parameterize model_name and temperature
    llm = AzureChatOpenAI(client=openai.ChatCompletion,
                            temperature=0, 
                            deployment_name="gpt-35-turbo",
                            model_name="gpt-35-turbo", 
                            openai_api_key=self.agent_connection.api_key,
                            openai_api_base=self.agent_connection.api_base,
                            openai_api_type=self.agent_connection.api_type,
                            openai_api_version=self.agent_connection.api_version)
    
    if self.memory_helper.use_history():
      print("Setting memory")
      memory = self.memory_helper.as_langchain_memory()
      print(f"Memory: {memory}")
      agent_type = AgentType.CONVERSATIONAL_REACT_DESCRIPTION
      agent = initialize_agent(
              tools,
              llm,
              agent=agent_type,
              memory=memory,
              verbose=True)
    else:
      agent_type = AgentType.ZERO_SHOT_REACT_DESCRIPTION
      agent = initialize_agent(
              tools,
              llm,
              agent=agent_type,
              verbose=True)
    answer = agent.run(query)
    return answer

  def use_openai_functions(self, query):
    # Call OpenAI to pick function
    print("Using Functions")
    functions = []
    function_names_mapping = {}
    for skill in self.skills:
      func_def = skill.to_function_definition()
      functions.append(func_def)
      function_names_mapping[func_def["name"]] = skill

    messages = [
      {
        "role": "system",
        "content": "Assistant is a helpful assistant that is good at using provided functions to get answers to specific questions. You must pick a function as a response"
      },
    ]

    if self.memory_helper.use_history():
      messages.extend(self.memory_helper.as_openai_function_memory())

    messages.append(
      {
        "role": "user",
        "content": query
      }
    )

    print(f"Functions: {functions}")
    response = openai.ChatCompletion.create(
        deployment_id="gpt-35-turbo",
        messages=messages,
        functions=functions,
        function_call="auto", 
    )
    response_message = response["choices"][0]["message"]
    print(f"Response from OpenAI: {response_message}")
    # Call Function
    if response_message.get("function_call"):
        print("Recommended Function call:")
        print(response_message.get("function_call"))
        print()
        
        # Step 3: call the function
        # Note: the JSON response may not always be valid; be sure to handle errors
        
        function_name = response_message["function_call"]["name"]
        
        # verify function exists
        if function_name not in function_names_mapping:
            return "Function " + function_name + " does not exist"
        tool_to_use = function_names_mapping[function_name]  
        
        # verify function has correct number of arguments
        function_args = json.loads(response_message["function_call"]["arguments"])
        print(f"Function Args: {function_args}")
        flow_result = tool_to_use.execute(function_args)

        print("Output of function call:")
        print(flow_result)
        print()

        messages.append(
            {
                "role": response_message["role"],
                "name": response_message["function_call"]["name"],
                "content": response_message["function_call"]["arguments"],
            }
        )

        # adding function response to messages
        messages.append(
            {
                "role": "function",
                "name": function_name,
                "content": json.dumps(flow_result),
            }
        )  # extend conversation with function response

        print("Messages in next request:")
        for message in messages:
            print(message)
        print()

        response = openai.ChatCompletion.create(
            messages=messages,
            deployment_id="gpt-35-turbo",
            function_call="none",
            functions=functions,
            temperature=0
        )

        answer = response['choices'][0]['message']["content"]
    else:
      answer = response_message

    return answer

  def use_semantic_kernel(self, query):
    my_logger = sk.NullLogger()
    kernel = sk.Kernel(log=my_logger)
    kernel.add_text_completion_service("dv", AzureChatCompletion("gpt-35-turbo", self.agent_connection.api_base, self.agent_connection.api_key))

    kernel.create_semantic_function(
        intent_prompt,
        description="Gets the intent of the user",
        function_name="GetIntent",
        skill_name="OrchestratorPlugin"
    )

    kernel.create_semantic_function(
      format_response,
      description="Takes a question and answer and formats it in a human readable way",
      function_name="FormatResponse",
      skill_name="OrchestratorPlugin"
    )


    # For each tool, import the skills (can we do this?)

    native_funcs = {}
    for skill in self.skills:
      print(f"Loading skill: {skill.name}")
      native_funcs[skill.name] = skill.to_sk_function()
    kernel.import_skill(native_funcs, "FlowPlugin")
    
    orchestratorPlugin = kernel.import_skill(
        OrchestratorPlugin(kernel), "OrchestratorPlugin"
    )
    print("Calling Orchestrator")
    
    if self.memory_helper.use_history():
      context = ContextVariables()
      # Use history to get intent
      kernel.create_semantic_function(
        intent_prompt_with_history,
        description="Gets the intent of the user",
        function_name="GetIntent",
        skill_name="OrchestratorPlugin"
      )

      # Add function to get final input for tool
      kernel.create_semantic_function(
        extract_input,
        description="Takes a question and using chat history figures out input for a plugin",
        function_name="ExtractInputs",
        skill_name="OrchestratorPlugin"
      )
      context["history"] = self.memory_helper.as_semantic_kernel_memory()
      context["input"] = query
      result = orchestratorPlugin["route_request"].invoke(
          query, context
      )
    else:
      kernel.create_semantic_function(
        intent_prompt,
        description="Gets the intent of the user",
        function_name="GetIntent",
        skill_name="OrchestratorPlugin"
      )
      result = orchestratorPlugin["route_request"].invoke(
          query
      )

    print(f"This is result: {result}")
    answer = result["input"]
    return answer


@tool
def pick_tool_and_execute(
    query: str,
    chat_history: list,
    arbitration_method: ArbitrationMethod,
    llm_connection: AzureOpenAIConnection,
    config1: AgentSkillConfiguration,
    config2: AgentSkillConfiguration,
    config3: AgentSkillConfiguration
):
  print(f"Arbitration Method: {arbitration_method}")
  print(f"Chat History: {chat_history}")
  print(f"Chat History Type: {type(chat_history)}")
  from importlib.metadata import version
  print(f"Langchain version: {version('langchain')}")  
  
  # Hacky way to get subscription info so we can communicate with WS
  deploy_config = PromptFlowRuntime.get_instance().config.deployment
  subscription_id = deploy_config.subscription_id
  resource_group_name=deploy_config.resource_group
  workspace_name=deploy_config.workspace_name

  # subscription_id = deploy_config.subscription_id if deploy_config.subscription_id != "" else "381b38e9-9840-4719-a5a0-61d9585e1e91"
  # resource_group_name=deploy_config.resource_group if deploy_config.resource_group != "" else "baker-static"
  # workspace_name=deploy_config.workspace_name if deploy_config.workspace_name != "" else "baker-eastus2"

  credential = DefaultAzureCredential()
  ml_client = MLClient(
      credential=credential,
      subscription_id=subscription_id,  # this will look like xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
      resource_group_name=resource_group_name,
      workspace_name=workspace_name,
  )
  pf = PFClient(ml_client) 
  print(f"Workspace Details: {subscription_id}, {resource_group_name}, {workspace_name}")

  agent_skills = []
  agent_skills.append(AgentSkill(config1, pf, subscription_id, resource_group_name, workspace_name))
  agent_skills.append(AgentSkill(config2, pf, subscription_id, resource_group_name, workspace_name))
  agent_skills.append(AgentSkill(config3, pf, subscription_id, resource_group_name, workspace_name))

  agent = ArbitrationAgent(arbitration_method, llm_connection, agent_skills, chat_history)

  return agent.execute(query)
